"""
This module contains tests for:

- test_kustomizations: Test if all kustomizations are ready
- test_helmreleases: Test if all the helmreleases are ready
- test_deployments: Test if all deployments are ready

Documentation for the kubernetes client:
* https://github.com/kubernetes-client/python/blob/master/kubernetes/README.md
* https://github.com/kubernetes-client/python/tree/master/examples
"""
import os
from kubernetes import client, config
from kubernetes.client.rest import ApiException

import pytest

# Helper functions

def check_resource(api, api_group, api_version, plural, name, namespace):
    """Returns status contition for resource in given namespace"""
    if not namespace:
        namespace='flux-system'
    print(f'Resource "{name}" in namespace "{namespace}" '
          f'(api version: {api_group}/{api_version}): ', end='')
    try:
        resource = api.get_namespaced_custom_object(
            group=api_group,
            version=api_version,
            plural=plural,
            name=name,
            namespace=namespace
        )
        status = resource['status']['conditions'][0]['status']
        print(status)
    except ApiException as ex:
        if ex.status == 404:
            status = 'Not found'
        else:
            raise
        print(f"**** NOT ready, status: {status} *****")
    except KeyError:
        status = 'Key error'
        print("resource key error: ")
        print(resource)
    return status

def check_custom_objects(api, api_group, api_version, plural, name, namespace='flux-system'):
    """
    Checks if all resoures of a given custom object are ready.

    We can't get a list of custom objects from all namespaces,
    so we have to iterate over all namespaces. See
    https://github.com/kubernetes-client/python/issues/1377
    """

    failed = 0
    failed_resources = []

    if name != 'all':
        status = check_resource(
            api=api,
            api_group=api_group,
            api_version=api_version,
            plural=plural,
            name=name,
            namespace=namespace)
        if status != 'True':
            failed += 1
            failed_resources.append(name)
    else:
        core_api = client.CoreV1Api()
        namespaces = core_api.list_namespace(watch=False)
        for namespace_object in namespaces.items:
            namespace_name = namespace_object.metadata.name
            custom_objects = api.list_namespaced_custom_object(
                group=api_group,
                version=api_version,
                plural=plural,
                namespace=namespace_name,
                watch=False
            )
            for custom_object in custom_objects['items']:
                resource_name = custom_object['metadata']['name']
                status = check_resource(
                    api=api,
                    api_group=api_group,
                    api_version=api_version,
                    plural=plural,
                    name=resource_name,
                    namespace=namespace_name)
                if status != 'True':
                    failed += 1
                    failed_resources.append(resource_name)
    assert failed == 0, f"Error: {failed} resources not ready ({failed_resources})!"

def check_all_pods_running(pods):
    """
    Loop through all the pods in an API result.
    If a pod does not have an element `status.phase` with the value "Running",
    return False. Otherwise, returns True.

    :param kubernetes.V1Pod[] pods: list of V1Pod elements to check
    """
    ret = True
    for pod in pods:
        print(f"- {pod.metadata.name}: {pod.status.phase}")
        if pod.status.phase != "Running":
            ret = False
    return ret


@pytest.fixture(autouse=True)
def run_around_tests():
    """
    Prepare kube config before running a test
    """
    cluster_dir = os.environ.get("CLUSTER_DIR")
    kubeconfig = os.path.join(cluster_dir, 'kube_config_cluster.yml')
    config.load_kube_config(config_file=kubeconfig)
    yield


# Pytest functions

@pytest.mark.resource
@pytest.mark.kustomizations
def test_kustomizations(resource, namespace):
    """
    Checks if all kustomizations installed by weave flux are in
    'Ready' state.
    """

    print('\n')
    custom_objects = client.CustomObjectsApi()
    check_custom_objects(
        api=custom_objects,
        api_group='kustomize.toolkit.fluxcd.io',
        api_version='v1beta1',
        plural='kustomizations',
        name=resource,
        namespace=namespace)


@pytest.mark.resource
@pytest.mark.helmreleases
def test_helmreleases(resource, namespace):
    """
    Checks if all HelmReleases installed by weave flux are in
    'Ready' state.
    """

    print('\n')
    custom_objects = client.CustomObjectsApi()
    check_custom_objects(
        api=custom_objects,
        api_group='helm.toolkit.fluxcd.io',
        api_version='v2beta1',
        plural='helmreleases',
        name=resource,
        namespace=namespace)


@pytest.mark.resource
@pytest.mark.deployments
def test_deployments(resource, namespace):
    """
    Check if deployments are ready
    """

    failed = 0
    failed_resources = []
    api = client.AppsV1Api()
    print('\n')

    if resource != 'all':
        deployments = api.list_namespaced_deployment(
            namespace=namespace,
            label_selector=f'helm.toolkit.fluxcd.io/name={resource}')
    else:
        deployments = api.list_deployment_for_all_namespaces()

    assert deployments.items, 'No deployments found !!'
    for deployment in deployments.items:
        resource = deployment.metadata.name
        namespace = deployment.metadata.namespace
        status = api.read_namespaced_deployment_status(resource, namespace)
        ready = status.status.conditions[0].status
        print(f'Deployment "{resource}" in namespace "{namespace}": {ready}')
        if ready != 'True':
            failed += 1
            failed_resources.append(resource)

    assert failed == 0, f"Error: {failed} resources not ready ({failed_resources})!"

@pytest.mark.resource
@pytest.mark.statefulsets
def test_statefulsets(resource, namespace):
    """
    Check if statefulsets are ready
    """

    failed = 0
    failed_resources = []
    api = client.AppsV1Api()
    print('\n')

    if resource != 'all':
        statefulsets = api.list_namespaced_stateful_set(
            namespace=namespace,
            label_selector=f'helm.toolkit.fluxcd.io/name={resource}')
    else:
        statefulsets = api.list_stateful_set_for_all_namespaces()

    assert statefulsets.items, 'No statefulsets found !!'
    for statefulset in statefulsets.items:
        resource = statefulset.metadata.name
        namespace = statefulset.metadata.namespace
        status = api.read_namespaced_stateful_set_status(resource, namespace)
        ready = status.status.replicas == status.status.ready_replicas
        print(f'statefulset "{resource}" in namespace "{namespace}": {ready}')
        if not ready:
            failed += 1
            failed_resources.append(resource)

    assert failed == 0, f"Error: {failed} resources not ready ({failed_resources})!"

@pytest.mark.resource
@pytest.mark.daemonsets
def test_daemonsets(resource, namespace):
    """
    Check if daemonsets are ready
    """

    failed = 0
    failed_resources = []
    api = client.AppsV1Api()
    print('\n')

    if resource != 'all':
        daemonsets = api.list_namespaced_daemon_set(
            namespace=namespace,
            label_selector=f'helm.toolkit.fluxcd.io/name={resource}')
    else:
        daemonsets = api.list_daemon_set_for_all_namespaces()

    assert daemonsets.items, 'No daemonsets found !!'
    for daemonset in daemonsets.items:
        resource = daemonset.metadata.name
        namespace = daemonset.metadata.namespace
        status = api.read_namespaced_daemon_set_status(resource, namespace)
        ready = status.status.desired_number_scheduled == status.status.number_available
        print(f'daemonset "{resource}" in namespace "{namespace}": {ready}')
        if not ready:
            failed += 1
            failed_resources.append(resource)

    assert failed == 0, f"Error: {failed} resources not ready ({failed_resources})!"
